@@ -4,18 +4,22 @@
 use {
     crate::{
         packet::{self, PacketBatch, PacketBatchRecycler, PACKETS_PER_BATCH},
+        recvmmsg::NUM_RCVMMSGS,
         sendmmsg::{batch_send, SendPktsError},
         socket::SocketAddrSpace,
     },
-    solana_sdk::timing::timestamp,
+    crossbeam_channel::{Receiver, RecvTimeoutError, SendError, Sender},
+    histogram::Histogram,
+    solana_sdk::{packet::Packet, timing::timestamp},
     std::{
-        net::UdpSocket,
+        cmp::Reverse,
+        collections::HashMap,
+        net::{IpAddr, UdpSocket},
         sync::{
-            atomic::{AtomicBool, AtomicUsize, Ordering},
-            mpsc::{Receiver, RecvTimeoutError, SendError, Sender},
+            atomic::{AtomicBool, Ordering},
             Arc,
         },
-        thread::{sleep, Builder, JoinHandle},
+        thread::{Builder, JoinHandle},
         time::{Duration, Instant},
     },
     thiserror::Error,
@@ -39,67 +43,24 @@ pub enum StreamerError {
     SendPktsError(#[from] SendPktsError),
 }
 
-pub struct StreamerReceiveStats {
-    pub name: &'static str,
-    pub packets_count: AtomicUsize,
-    pub packet_batches_count: AtomicUsize,
-    pub full_packet_batches_count: AtomicUsize,
-    pub max_channel_len: AtomicUsize,
-}
-
-impl StreamerReceiveStats {
-    pub fn new(name: &'static str) -> Self {
-        Self {
-            name,
-            packets_count: AtomicUsize::default(),
-            packet_batches_count: AtomicUsize::default(),
-            full_packet_batches_count: AtomicUsize::default(),
-            max_channel_len: AtomicUsize::default(),
-        }
-    }
-
-    pub fn report(&self) {
-        datapoint_info!(
-            self.name,
-            (
-                "packets_count",
-                self.packets_count.swap(0, Ordering::Relaxed) as i64,
-                i64
-            ),
-            (
-                "packet_batches_count",
-                self.packet_batches_count.swap(0, Ordering::Relaxed) as i64,
-                i64
-            ),
-            (
-                "full_packet_batches_count",
-                self.full_packet_batches_count.swap(0, Ordering::Relaxed) as i64,
-                i64
-            ),
-            (
-                "channel_len",
-                self.max_channel_len.swap(0, Ordering::Relaxed) as i64,
-                i64
-            ),
-        );
-    }
-}
-
 pub type Result<T> = std::result::Result<T, StreamerError>;
 
 fn recv_loop(
-    socket: &UdpSocket,
+    sock: &UdpSocket,
     exit: Arc<AtomicBool>,
-    packet_batch_sender: &PacketBatchSender,
+    channel: &PacketBatchSender,
     recycler: &PacketBatchRecycler,
-    stats: &StreamerReceiveStats,
+    name: &'static str,
     coalesce_ms: u64,
     use_pinned_memory: bool,
-    in_vote_only_mode: Option<Arc<AtomicBool>>,
 ) -> Result<()> {
+    let mut recv_count = 0;
+    let mut call_count = 0;
+    let mut now = Instant::now();
+    let mut num_max_received = 0; // Number of times maximum packets were received
     loop {
         let mut packet_batch = if use_pinned_memory {
-            PacketBatch::new_with_recycler(recycler.clone(), PACKETS_PER_BATCH, stats.name)
+            PacketBatch::new_with_recycler(recycler.clone(), PACKETS_PER_BATCH, name)
         } else {
             PacketBatch::with_capacity(PACKETS_PER_BATCH)
         };
@@ -109,73 +70,182 @@ fn recv_loop(
             if exit.load(Ordering::Relaxed) {
                 return Ok(());
             }
-
-            if let Some(ref in_vote_only_mode) = in_vote_only_mode {
-                if in_vote_only_mode.load(Ordering::Relaxed) {
-                    sleep(Duration::from_millis(1));
-                    continue;
+            if let Ok(len) = packet::recv_from(&mut packet_batch, sock, coalesce_ms) {
+                if len == NUM_RCVMMSGS {
+                    num_max_received += 1;
                 }
-            }
-
-            if let Ok(len) = packet::recv_from(&mut packet_batch, socket, coalesce_ms) {
+                recv_count += len;
+                call_count += 1;
                 if len > 0 {
-                    let StreamerReceiveStats {
-                        packets_count,
-                        packet_batches_count,
-                        full_packet_batches_count,
-                        ..
-                    } = stats;
-
-                    packets_count.fetch_add(len, Ordering::Relaxed);
-                    packet_batches_count.fetch_add(1, Ordering::Relaxed);
-                    if len == PACKETS_PER_BATCH {
-                        full_packet_batches_count.fetch_add(1, Ordering::Relaxed);
-                    }
-
-                    packet_batch_sender.send(packet_batch)?;
+                    channel.send(packet_batch)?;
                 }
                 break;
             }
         }
+        if recv_count > 1024 {
+            datapoint_debug!(
+                name,
+                ("received", recv_count as i64, i64),
+                ("call_count", i64::from(call_count), i64),
+                ("elapsed", now.elapsed().as_millis() as i64, i64),
+                ("max_received", i64::from(num_max_received), i64),
+            );
+            recv_count = 0;
+            call_count = 0;
+            num_max_received = 0;
+        }
+        now = Instant::now();
     }
 }
 
 pub fn receiver(
-    socket: Arc<UdpSocket>,
-    exit: Arc<AtomicBool>,
-    packet_batch_sender: PacketBatchSender,
+    sock: Arc<UdpSocket>,
+    exit: &Arc<AtomicBool>,
+    packet_sender: PacketBatchSender,
     recycler: PacketBatchRecycler,
-    stats: Arc<StreamerReceiveStats>,
+    name: &'static str,
     coalesce_ms: u64,
     use_pinned_memory: bool,
-    in_vote_only_mode: Option<Arc<AtomicBool>>,
 ) -> JoinHandle<()> {
-    let res = socket.set_read_timeout(Some(Duration::new(1, 0)));
+    let res = sock.set_read_timeout(Some(Duration::new(1, 0)));
     assert!(res.is_ok(), "streamer::receiver set_read_timeout error");
+    let exit = exit.clone();
     Builder::new()
         .name("solana-receiver".to_string())
         .spawn(move || {
             let _ = recv_loop(
-                &socket,
+                &sock,
                 exit,
-                &packet_batch_sender,
-                &recycler,
-                &stats,
+                &packet_sender,
+                &recycler.clone(),
+                name,
                 coalesce_ms,
                 use_pinned_memory,
-                in_vote_only_mode,
             );
         })
         .unwrap()
 }
 
+#[derive(Debug, Default)]
+struct SendStats {
+    bytes: u64,
+    count: u64,
+}
+
+#[derive(Default)]
+struct StreamerSendStats {
+    host_map: HashMap<IpAddr, SendStats>,
+    since: Option<Instant>,
+}
+
+impl StreamerSendStats {
+    fn report_stats(
+        name: &'static str,
+        host_map: HashMap<IpAddr, SendStats>,
+        sample_duration: Option<Duration>,
+    ) {
+        const MAX_REPORT_ENTRIES: usize = 5;
+        let sample_ms = sample_duration.map(|d| d.as_millis()).unwrap_or_default();
+        let mut hist = Histogram::default();
+        let mut byte_sum = 0;
+        let mut pkt_count = 0;
+        host_map.iter().for_each(|(_addr, host_stats)| {
+            hist.increment(host_stats.bytes).unwrap();
+            byte_sum += host_stats.bytes;
+            pkt_count += host_stats.count;
+        });
+
+        datapoint_info!(
+            name,
+            ("streamer-send-sample_duration_ms", sample_ms, i64),
+            ("streamer-send-host_count", host_map.len(), i64),
+            ("streamer-send-bytes_total", byte_sum, i64),
+            ("streamer-send-pkt_count_total", pkt_count, i64),
+            (
+                "streamer-send-host_bytes_min",
+                hist.minimum().unwrap_or_default(),
+                i64
+            ),
+            (
+                "streamer-send-host_bytes_max",
+                hist.maximum().unwrap_or_default(),
+                i64
+            ),
+            (
+                "streamer-send-host_bytes_mean",
+                hist.mean().unwrap_or_default(),
+                i64
+            ),
+            (
+                "streamer-send-host_bytes_90pct",
+                hist.percentile(90.0).unwrap_or_default(),
+                i64
+            ),
+            (
+                "streamer-send-host_bytes_50pct",
+                hist.percentile(50.0).unwrap_or_default(),
+                i64
+            ),
+            (
+                "streamer-send-host_bytes_10pct",
+                hist.percentile(10.0).unwrap_or_default(),
+                i64
+            ),
+        );
+
+        let num_entries = host_map.len();
+        let mut entries: Vec<_> = host_map.into_iter().collect();
+        if entries.len() > MAX_REPORT_ENTRIES {
+            entries.select_nth_unstable_by_key(MAX_REPORT_ENTRIES, |(_addr, stats)| {
+                Reverse(stats.bytes)
+            });
+            entries.truncate(MAX_REPORT_ENTRIES);
+        }
+        info!(
+            "streamer send {} hosts: count:{} {:?}",
+            name, num_entries, entries,
+        );
+    }
+
+    fn maybe_submit(&mut self, name: &'static str, sender: &Sender<Box<dyn FnOnce() + Send>>) {
+        const SUBMIT_CADENCE: Duration = Duration::from_secs(10);
+        const MAP_SIZE_REPORTING_THRESHOLD: usize = 1_000;
+        let elapsed = self.since.as_ref().map(Instant::elapsed);
+        if elapsed.map(|e| e < SUBMIT_CADENCE).unwrap_or_default()
+            && self.host_map.len() < MAP_SIZE_REPORTING_THRESHOLD
+        {
+            return;
+        }
+
+        let host_map = std::mem::take(&mut self.host_map);
+        let _ = sender.send(Box::new(move || {
+            Self::report_stats(name, host_map, elapsed);
+        }));
+
+        *self = Self {
+            since: Some(Instant::now()),
+            ..Self::default()
+        };
+    }
+
+    fn record(&mut self, pkt: &Packet) {
+        let ent = self.host_map.entry(pkt.meta.addr).or_default();
+        ent.count += 1;
+        ent.bytes += pkt.data.len() as u64;
+    }
+}
+
 fn recv_send(
     sock: &UdpSocket,
     r: &PacketBatchReceiver,
     socket_addr_space: &SocketAddrSpace,
+    stats: &mut Option<StreamerSendStats>,
 ) -> Result<()> {
     let timer = Duration::new(1, 0);
     let packet_batch = r.recv_timeout(timer)?;
+    if let Some(stats) = stats {
+        packet_batch.packets.iter().for_each(|p| stats.record(p));
+    }
     let packets = packet_batch.packets.iter().filter_map(|pkt| {
         let addr = pkt.meta.addr();
         socket_addr_space
@@ -214,6 +284,7 @@ pub fn responder(
     sock: Arc<UdpSocket>,
     r: PacketBatchReceiver,
     socket_addr_space: SocketAddrSpace,
+    stats_reporter_sender: Option<Sender<Box<dyn FnOnce() + Send>>>,
 ) -> JoinHandle<()> {
     Builder::new()
         .name(format!("solana-responder-{}", name))
@@ -221,8 +292,14 @@ pub fn responder(
             let mut errors = 0;
             let mut last_error = None;
             let mut last_print = 0;
+            let mut stats = None;
+
+            if stats_reporter_sender.is_some() {
+                stats = Some(StreamerSendStats::default());
+            }
+
             loop {
-                if let Err(e) = recv_send(&sock, &r, &socket_addr_space) {
+                if let Err(e) = recv_send(&sock, &r, &socket_addr_space, &mut stats) {
                     match e {
                         StreamerError::RecvTimeout(RecvTimeoutError::Disconnected) => break,
                         StreamerError::RecvTimeout(RecvTimeoutError::Timeout) => (),
@@ -239,6 +316,11 @@ pub fn responder(
                     last_print = now;
                     errors = 0;
                 }
+                if let Some(ref stats_reporter_sender) = stats_reporter_sender {
+                    if let Some(ref mut stats) = stats {
+                        stats.maybe_submit(name, stats_reporter_sender);
+                    }
+                }
             }
         })
         .unwrap()
@@ -252,6 +334,7 @@ mod test {
             packet::{Packet, PacketBatch, PACKET_DATA_SIZE},
             streamer::{receiver, responder},
         },
+        crossbeam_channel::unbounded,
         solana_perf::recycler::Recycler,
         std::{
             io,
@@ -259,7 +342,6 @@ mod test {
             net::UdpSocket,
             sync::{
                 atomic::{AtomicBool, Ordering},
-                mpsc::channel,
                 Arc,
             },
             time::Duration,
@@ -294,29 +376,27 @@ mod test {
         let addr = read.local_addr().unwrap();
         let send = UdpSocket::bind("127.0.0.1:0").expect("bind");
         let exit = Arc::new(AtomicBool::new(false));
-        let (s_reader, r_reader) = channel();
-        let stats = Arc::new(StreamerReceiveStats::new("test"));
+        let (s_reader, r_reader) = unbounded();
         let t_receiver = receiver(
             Arc::new(read),
-            exit.clone(),
+            &exit,
             s_reader,
             Recycler::default(),
-            stats.clone(),
+            "test",
             1,
             true,
-            None,
         );
-        const NUM_PACKETS: usize = 5;
         let t_responder = {
-            let (s_responder, r_responder) = channel();
+            let (s_responder, r_responder) = unbounded();
             let t_responder = responder(
                 "streamer_send_test",
                 Arc::new(send),
                 r_responder,
                 SocketAddrSpace::Unspecified,
+                None,
             );
             let mut packet_batch = PacketBatch::default();
-            for i in 0..NUM_PACKETS {
+            for i in 0..5 {
                 let mut p = Packet::default();
                 {
                     p.data[0] = i as u8;
@@ -329,13 +409,10 @@ mod test {
             t_responder
         };
 
-        let mut packets_remaining = NUM_PACKETS;
+        let mut packets_remaining = 5;
         get_packet_batches(r_reader, &mut packets_remaining);
         assert_eq!(packets_remaining, 0);
         exit.store(true, Ordering::Relaxed);
-        assert_eq!(stats.packets_count.load(Ordering::Relaxed), NUM_PACKETS);
-        assert_eq!(stats.packet_batches_count.load(Ordering::Relaxed), 1);
-        assert_eq!(stats.full_packet_batches_count.load(Ordering::Relaxed), 0);
         t_receiver.join().expect("join");
         t_responder.join().expect("join");
     }
